#![allow(unused, missing_docs)] // This file is for demo purposes
//! Examples of using SNAFU
//!
//! You are encouraged to view the Rust source for each error type by
//! clicking the `[src]` link in the upper-right of each documentation
//! page. This shows how you would accomplish the same task in your
//! code.
//!
//! The suggested order to browse them is:
//!
//! 1. [`basic`]
//! 1. [`backtrace`]
//!
//! You should also feel free to browse the [acceptance tests][tests]
//! and [compatibility tests][compat], which cover a broad range of
//! functionality but with minimal descriptive prose.
//!
//! [tests]: https://github.com/shepmaster/snafu/tree/master/tests
//! [compat]: https://github.com/shepmaster/snafu/tree/master/compatibility-tests

pub mod backtrace;
pub mod basic;
